library out_networkserver;

(* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version 
 * 1.1 (the "License"); you may not use this file except in compliance with 
 * the License. You may obtain a copy of the License at 
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Network Client Output Module for Lightning.
 *
 * The Initial Developer of the Original Code is
 * Tomsoft.
 * Portions created by Tomsoft are Copyright (C) 2012
 * Tomsoft. All Rights Reserved.
 *
 * Contributor(s):
 *   Tommy Sools <tom@sools.com>
 *
 * ***** END LICENSE BLOCK ***** *)

{$IMAGEBASE $41230000}

uses
  SysUtils,
  Windows,
  StdCtrls,
  Graphics,
  WinSock,
  common in '..\common.pas',
  WSockets in '..\components\WSockets.pas';

{$R *.RES}

const
  channels = 32;
  defaultPort = 80055;

type
  TPacket = record
    channel: 1..channels;
    level:   TLevel;
  end;
  TDetails = class(TObject)
  public
    gbDetails:   TGroupBox;
    lPort:       TLabel;
    ePort:       TEdit;
    bUpdate:     TButton;
    stConnected: TStaticText;
    chBinary:    TCheckBox;
    bTCPServer:  TTCPServer;
    levels:      array[1..channels] of TLevel;
    status:      array[1..channels] of Boolean;
    binary:      Boolean;
    lastInt:     Integer;
    procedure bUpdateClick(Sender: TObject);
    procedure chBinaryClick(Sender: TObject);
    procedure tcpServerAccept(Sender: TObject; Socket: TSocket);
    procedure updateConnectionState;
  end;

procedure TDetails.bUpdateClick(Sender: TObject);
begin
  bTCPServer.Port := ePort.Text;
  if bTCPServer.SocketState <> ssClosed then
    bTCPServer.Close();
  bTCPServer.Open();
  updateConnectionState();
end;

procedure TDetails.chBinaryClick(Sender: TObject);
begin
  binary := chBinary.Checked;
end;

procedure TDetails.tcpServerAccept(Sender: TObject; Socket: TSocket);
var
  i, int, check: Integer;
  packet: TPacket;
begin
  if binary then
  begin
    int   := 0;
    check := 0;
    for i := 1 to channels do
    begin
      if i = 1 then
        check := 1
      else
        check := check * 2;
      if status[i] then
        int := int or check;
    end;
    bTCPServer.WriteBuffer(Socket, @int, SizeOf(int));
  end
  else
    for i := 1 to channels do
    begin
      packet.level  := levels[i];
      packet.channel := i;
      bTCPServer.WriteBuffer(Socket, @packet, SizeOf(packet))
    end;
end;

procedure TDetails.updateConnectionState;
begin
  if gbDetails <> nil then
    if bTCPServer.SocketState <> ssListening then
    begin
      stConnected.Font.Color := clRed;
      stConnected.Caption    := 'Not listening.';
    end
    else
    begin
      stConnected.Font.Color := clGreen;
      stConnected.Caption    := 'Listening.';
    end;
end;

function  init(name: PChar; nameLength: DWORD): DWORD; cdecl;
begin
  Result := return('ONetworkServer', name, nameLength);
end;

procedure finish(); cdecl;
begin
  //
end;

function  create(controlHandle: THandle): PLightningModule; cdecl;
var
  myDetails: TDetails;
  i:         Integer;
begin
  New(Result);
  with Result^ do
  begin
    handle        := controlHandle;
    title         := 'NetworkServer';
    settings      := True;
    width         := -1;
    height        := -1;
    inputTop      := -1;
    infoTop       := -1;
    outputTop     := -1;
    inputs        := channels;
    outputs       := 0;
    GetMem(inputTooltip, maxLength);
    StrCopy(inputTooltip, '');
    for i := 1 to channels do
      StrCat(inputTooltip, PChar('Channel ' + IntToStr(i) + ' value|'));
    outputTooltip := '';
    myDetails     := TDetails.Create();
    details       := myDetails;
    reserved      := nil;
  end;
  with myDetails do
  begin
    binary := False;
    for i := 1 to channels do
    begin
      levels[i] := levelMin;
      status[i] := False;
    end;
    lastInt := 0;
    gbDetails := nil;
    bTCPServer          := TTCPServer.Create(nil);
    bTCPServer.OnAccept := tcpServerAccept;
    bTCPServer.Port     := IntToStr(defaultPort);
  end;
end;

function  action(lightningModule: PLightningModule; input, output: Integer; inputConnected, outputConnected: PChar; var bridge: TSpectrum): Boolean; cdecl;
var
  details: TDetails;
  packet:  TPacket;
  bString: string;
  i, int, check, lastConnected: Integer;
begin
  Result  := False;
  details := TDetails(lightningModule^.details);
 if details.bTCPServer.SocketState = ssListening then 
  if details.binary then
  begin
    details.status[input] := spectrumAverage(bridge, fftMin, fftMax) > enabledBoundary;
    lastConnected := channels + 1;
    bString := inputConnected;
    while Copy(bString, Length(bString), 1) = '|' do
    begin
      Delete(bString, Length(bString), 1);
      Dec(lastConnected);
    end;
    if input = lastConnected then // if this is the last input that's connected
    begin
      int   := 0;
      check := 0;
      for i := 1 to channels do
      begin
        if i = 1 then
          check := 1
        else
          check := check * 2;
        if details.status[i] then
          int := int or check;
      end;
      if int <> details.lastInt then
      begin
        for i := 0 to details.bTCPServer.Clients.Count - 1 do
          details.bTCPServer.WriteBuffer(details.bTCPServer.Clients.Sockets[i], @int, SizeOf(int));
        details.lastInt := int;
      end;
    end;
  end
  else
  begin
    packet.level := spectrumAverage(bridge, fftMin, fftMax);
    if packet.level <> details.levels[input] then
    begin
      details.levels[input] := packet.level;
      packet.channel := input;
      for i := 0 to details.bTCPServer.Clients.Count - 1 do
        details.bTCPServer.WriteBuffer(details.bTCPServer.Clients.Sockets[i], @packet, SizeOf(packet))
    end;
  end;
end;

procedure show(lightningModule: PLightningModule; handle: THandle); cdecl;
begin
  with TDetails(lightningModule^.details) do
  begin
    gbDetails         := TGroupBox.Create(nil);
    gbDetails.Top     := detailsTop;
    gbDetails.Left    := detailsLeft;
    gbDetails.Width   := detailsWidth;
    gbDetails.Height  := detailsHeight;
    gbDetails.Ctl3D   := True;
    gbDetails.Color   := clBtnFace;
    gbDetails.Caption := 'Network Server Settings';
    lPort := TLabel.Create(gbDetails);
    with lPort do
    begin
      AutoSize := True;
      Top      := 20;
      Left     := 8;
      Caption  := 'Server port:';
      Parent   := gbDetails;
    end;
    ePort := TEdit.Create(gbDetails);
    with ePort do
    begin
      Top    := 40;
      Left   := 8;
      Width  := 120;
      Text   := bTCPServer.Port;
      Parent := gbDetails;
    end;
    bUpdate := TButton.Create(gbDetails);
    with bUpdate do
    begin
      Top     := 40;
      Left    := 140;
      Width   := 120;
      Height  := ePort.Height;
      Caption := 'Update';
      OnClick := bUpdateClick;
      Parent  := gbDetails;
    end;
    stConnected := TStaticText.Create(gbDetails);
    with stConnected do
    begin
      AutoSize   := True;
      Top        := 20;
      Left       := 140;
      Font.Style := [fsBold];
      updateConnectionState;
      Parent     := gbDetails;
    end;
    chBinary := TCheckBox.Create(gbDetails);
    with chBinary do
    begin
      Top     := 70;
      Left    := 8;
      Width   := 300;
      Caption := 'Binary mode (for better performance)';
      Checked := binary;
      OnClick := chBinaryClick;
      Parent  := gbDetails;
    end;
    gbDetails.ParentWindow := handle;
  end;
end;

procedure hide(lightningModule: PLightningModule); cdecl;
begin
  with TDetails(lightningModule^.details) do
  begin
    chBinary.Free();
    stConnected.Free();
    bUpdate.Free();
    ePort.Free();
    lPort.Free();
    gbDetails.Free();
    gbDetails := nil;
  end;
end;

procedure open(lightningModule: PLightningModule; settings: PChar); cdecl;
var
  bString, bString2: string;
begin
  bString := settings;
  with TDetails(lightningModule^.details) do
  begin
    parse(bString, ',', bString2);
    if Trim(bString2) = '' then
      bString2 := IntToStr(defaultPort);
    bTCPServer.Port := bString2;
    binary := parse(bString, ',') <> '';
    if bTCPServer.SocketState <> ssClosed then
      bTCPServer.Close();
    bTCPServer.Open();
    updateConnectionState();
  end;
end;

function  save(lightningModule: PLightningModule; settings: PChar; settingsLength: DWORD): DWORD; cdecl;
var
  bString: string;
begin
  with TDetails(lightningModule^.details) do
  begin
    bString := bTCPServer.Port + ',';
    if binary then
      bString := bString + '1';
    Result := return(bString, settings, settingsLength);
  end;
end;

procedure destroy(lightningModule: PLightningModule); cdecl;
begin
  FreeMem(lightningModule^.inputTooltip);
  TDetails(lightningModule^.details).bTCPServer.Free;
  Dispose(lightningModule);
end;

exports
  init, finish, create, destroy, action, show, hide, open, save;

begin
end.
